-- This Oracle SQL*Plus script can be used to check if the instance 
--   - Is RAC enabled
--   - Has TAF failover configured
--   - Has FAN events are enabled
-- using a 10g server or greater.
--
-- For Cache Connect to take full advantage of Oracle RAC, the attributes 
-- should be:
--   RAC Enabled:   YES
--   Failover Type: SELECT
--   FAN Events:    YES
--
-- Note: 
--   In order to use this script you must have SELECT privileges on 
--   sys.v_$option, sys.v_$session, and sys.v_$services.
--

SET SERVEROUTPUT ON;

DECLARE 
  racEnabled      varchar(7);
  failoverType    v$session.failover_type%TYPE := 'NONE';
  fanNotification varchar(3)                   := 'NO';
BEGIN
  SELECT CASE value 
         WHEN 'TRUE' THEN 'YES' 
         WHEN 'FALSE' THEN 'NO'
         ELSE 'UNKNOWN' END
  INTO racEnabled
  FROM v$option 
  WHERE parameter='Real Application Clusters';

  IF (racEnabled = 'YES') THEN
    SELECT UPPER(failover_type)
    INTO failoverType 
    FROM  v$session 
    WHERE sys_context('USERENV','SID') = sid;

    SELECT UPPER(aq_ha_notification)
    INTO fanNotification
    FROM v$services 
    WHERE sys_context('USERENV', 'SERVICE_NAME') 
            = nvl2(SYS_CONTEXT('USERENV','DB_DOMAIN'), 
                   name || '.' || SYS_CONTEXT('USERENV','DB_DOMAIN'), 
                   name);
  END IF;

  DBMS_OUTPUT.PUT_LINE('RAC Enabled:   ' || racEnabled);
  DBMS_OUTPUT.PUT_LINE('Failover Type: ' || failoverType);
  DBMS_OUTPUT.PUT_LINE('FAN Events:    ' || fanNotification);
END;
.
/

SET SERVEROUTPUT OFF;

